%{ open Ast %}

%token SEMI LPAREN RPAREN LBRACE RBRACE LBRACK RBRACK COMMA COLON DOT
%token PLUS MINUS TIMES DIVIDE MOD
%token EQ NEQ LT LEQ GT GEQ
%token INT_TYPE FLOAT_TYPE STRING_TYPE BOOLEAN_TYPE
%token ARROW COMMA
%token OBJECT_DEF OBJECT_TYPE 
%token TABLE_DEF TABLE_TYPE
%token CSV_DEF
%token IF THEN ELSE
%token CSV_INPUT
%token FOR FOREACH
%token RETURN
%token FUNC
%token ASSIGN
%token <string> STRING_LITERAL
%token <float> FLOAT
%token <int> LITERAL
%token <string> ID
%token <bool> BOOLLIT
%token EOF

%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD

%start program
%type <Ast.program> program

%%

program:
   /* nothing */               { { varDecls = []; funcDecls = []; objDecls = []; tableDecls = []; stmts = []; csvDecls = [] } }
  | program var_decl           { { varDecls = $2 :: $1.varDecls; funcDecls = $1.funcDecls; objDecls = $1.objDecls; tableDecls = $1.tableDecls; stmts = $1.stmts; csvDecls = $1.csvDecls } }
  | program func_decl          { { varDecls = $1.varDecls; funcDecls = $2 :: $1.funcDecls; objDecls = $1.objDecls; tableDecls = $1.tableDecls; stmts = $1.stmts; csvDecls = $1.csvDecls } }
  | program obj_decl           { { varDecls = $1.varDecls; funcDecls = $1.funcDecls; objDecls = $2 :: $1.objDecls; tableDecls = $1.tableDecls; stmts = $1.stmts; csvDecls = $1.csvDecls } }
  | program table_decl         { { varDecls = $1.varDecls; funcDecls = $1.funcDecls; objDecls = $1.objDecls; tableDecls = $2 :: $1.tableDecls; stmts = $1.stmts; csvDecls = $1.csvDecls } }
  | program stmt               { { varDecls = $1.varDecls; funcDecls = $1.funcDecls; objDecls = $1.objDecls; tableDecls = $1.tableDecls; stmts = $2 :: $1.stmts; csvDecls = $1.csvDecls } }
  | program csv_decl         { { varDecls = $1.varDecls; funcDecls = $1.funcDecls; objDecls = $1.objDecls; tableDecls = $1.tableDecls; stmts = $1.stmts; csvDecls = $2 :: $1.csvDecls } }

obj_decl:
  OBJECT_DEF ID LBRACE field_list RBRACE SEMI { { objName = $2; objFields = List.rev $4 } }

table_decl:
  TABLE_DEF ID LPAREN ID RPAREN LBRACE table_column_list RBRACE SEMI { { tableName = $2; iterClass = $4; columns = List.rev $7 } }

csv_decl:
  CSV_DEF LPAREN ID RPAREN SEMI                { { csvInputClass = $3 } }

table_column_list:
  table_column                                 { [$1] }
  | table_column_list COMMA table_column       { $3 :: $1 }  

table_column:
  ID ARROW type_name COLON column_func     {{ tblColName = $1; tblColType = $3; tblColFunc = $5 }}

column_func:
   var_decl_list stmt_list expr            { { locals = List.rev $1; statements = List.rev $2; lastExpr = $3 } }

field_list:
  ID ARROW type_name           { [ ( $1, $3 ) ] }
  | field_list COMMA ID ARROW type_name { ($3, $5) :: $1 }

var_decl:
  type_name var_decl_id assign_opt SEMI { { vname = fst $2; varType = $1; varArraySize = snd $2; varAssign = $3 } }

var_decl_id:
  ID array_size_opt           { ($1,$2) }

array_size_opt:
  /*  nothing */               { None }
 | LBRACK LITERAL RBRACK       { Some($2) }

func_var_decl:
  type_name ID                 { { vname = $2; varType = $1; varArraySize = None; varAssign = None } }

assign_opt:
  /* nothing */                { None }
  |  ASSIGN initialize_assign  { Some($2) }

initialize_assign:                   
  | constructor_call           { [$1] }
  | init_item                  { [$1] }
  | array_init                 { $1 }

array_init:
  | LBRACE array_init_list RBRACE { List.rev $2 }

array_init_list:
  | init_item                       { [$1] }
  | array_init_list COMMA init_item { $3 :: $1 }

init_item:
  constant_expr                { $1 }

type_name:
    INT_TYPE                   { Int }
  | FLOAT_TYPE                 { Float }
  | STRING_TYPE                { String }
  | BOOLEAN_TYPE               { Boolean }
  | OBJECT_TYPE ID             { Record($2) }
  | TABLE_TYPE ID              { Table($2) }
/*  | ID                         { Record($1) }  */

func_decl:
  FUNC type_name ID LPAREN params_opt RPAREN LBRACE var_decl_list stmt_list RBRACE
    { { name = $3; formals = $5; locals = List.rev $8; body = List.rev $9; retType = $2 } }

params_opt:
    /* nothing */              { [] }
  | params_list                { List.rev $1 } 

params_list:
    func_var_decl              { [$1] }
  | params_list COMMA func_var_decl { $3 :: $1 }

var_decl_list:
    /* Nothing */                { [] }
  | var_decl_list var_decl       { $2 :: $1 }

stmt_list:
    /* nothing  */               { [] }
  | stmt_list stmt               { $2 :: $1 }

stmt:
    expr SEMI                    { Expr($1) }
  | FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN LBRACE stmt_list RBRACE { For( $3, $5, $7, List.rev $10 ) }
  | FOREACH LPAREN ID RPAREN LBRACE stmt_list RBRACE { Foreach($3,List.rev $6) } 
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | RETURN expr SEMI             { Return($2) }

expr_opt:
  /* nothing */                  { Noexpr }
  | expr                         { $1 }

constant_expr:
    LITERAL                      { Literal($1) }
  | FLOAT                        { LitFloat($1) }
  | BOOLLIT                      { LitBool($1) }
  | STRING_LITERAL               { LitString($1) }

constructor_call:
  | ID LPAREN args_opt RPAREN    { Call( $1, $3 ) }

expr:
  constant_expr                  { $1 }
  | ID                           { Id($1) }
  | CSV_INPUT                    { CSVInput }
  | ID DOT ID                    { MemberAccess($1,$3) }
  | ID DOT ID LPAREN args_opt RPAREN { MethodCall($1,$3,$5) }
  | ID LBRACK expr RBRACK        { ArrayAccess($1,$3)  }
  | expr PLUS expr               { Binop( $1, Add, $3 ) }
  | expr MINUS expr              { Binop( $1, Sub, $3 ) }
  | expr TIMES expr              { Binop( $1, Mult, $3 ) }
  | expr DIVIDE expr             { Binop( $1, Div, $3 ) }
  | expr MOD expr                { Binop( $1, Mod, $3 ) }
  | expr EQ     expr             { Binop($1, Equal, $3) }
  | expr NEQ    expr             { Binop($1, Neq,   $3) }
  | expr LT     expr             { Binop($1, Less,  $3) }
  | expr LEQ    expr             { Binop($1, Leq,   $3) }
  | expr GT     expr             { Binop($1, Greater,  $3) }
  | expr GEQ    expr             { Binop($1, Geq,   $3) }
  | expr ASSIGN expr             { Assign($1,$3) }
  | IF LPAREN expr RPAREN  THEN LBRACE stmt_list expr RBRACE 
    ELSE LBRACE stmt_list expr RBRACE      { IfThenElse( $3, $8, $13, $7, $12) }
  | constructor_call             { $1 }

args_opt:
    /* nothing */                { [] }
  | args_list                    { List.rev $1 }

args_list:
  | expr                         { [$1] }
  | args_list COMMA expr         { $3 :: $1 }


    
  




  